# Module to save and load data in Perl.
# Examples:
# use std;
# my $data = std::readfile('file.txt');
# my $data = std::readfile_binmode('image.jpg');
# std::savefile('file.txt',$data);
# std::savefile_binmode('file.txt',$data);
# my @lines = std::readfile_array_nobr('file.txt');
# std::xwrite("databasename","id",$data); 
# (is saved in ./equals/databasename/hashid.txt)
# my $data = std::xread("databasename","id"); 
# you can also load and save arrays in xread and xwrite
# std::xwrite("databasename","id",@array); 
# my @array = std::xread("databasename","id");
package std;


#------------ uses
use LWP::UserAgent;
use Digest::SHA; #old-> Digest::SHA1  qw(sha1 sha1_hex sha1_base64);
#use Compress::Bzip2;
#use Compress::Zlib;
use Fcntl ':flock';

#==== einzigartige elemente einer liste
sub uniq { return keys %{{ map { $_ => 1 } @_ }}; }

#===================================== STANDARD SUBS
sub opendir($)
{
 if(opendir(DIR, $_[0]))
 {
  my @dirs = readdir(DIR);
  closedir(DIR);
  return(@dirs);
 }
}
sub opendir_files($)
{
 my $dir = $_[0];
 $dir =~ s/\\/\//g;
 $dir = (substr($dir,-1) ne "/" && $dir ne "")? $dir."/": $dir;
 if(opendir(DIR, $dir))
 {
  my @dirs = readdir(DIR);
  closedir(DIR);
  return(grep(-f $dir.$_,@dirs));
 }
}
sub opendir_pathfiles($)
{
 my $dir = $_[0];
 $dir =~ s/\\/\//g;
 $dir = (substr($dir,-1) ne "/" && $dir ne "")? $dir."/": $dir;
 if(opendir(DIR, $dir))
 {
  my @dirs = readdir(DIR);
  closedir(DIR);
  return(grep(-f $_,map {$dir.$_;} @dirs));
 }
}
sub opendir_dirs($)
{
 my $dir = $_[0];
 $dir =~ s/\\/\//g;
 $dir = (substr($dir,-1) ne "/" && $dir ne "")? $dir."/": $dir;
 if(opendir(DIR, $dir))
 {
  my @dirs = readdir(DIR);
  closedir(DIR);
  return(grep(-d $dir.$_ && $_ ne "." && $_ ne "..",@dirs));
 }
}
sub opendir_pathdirs($)
{
 my $dir = $_[0];
 $dir =~ s/\\/\//g;
 $dir = (substr($dir,-1) ne "/" && $dir ne "")? $dir."/": $dir;
 if(opendir(DIR, $dir))
 {
  my @dirs = readdir(DIR);
  closedir(DIR);
  return(map {$dir.$_} grep(-d $dir.$_ && $_ ne "." && $_ ne "..",@dirs));
 }
}
sub opendir_subpathdirs($)
{
 my $dir = $_[0];
 $dir =~ s/\\/\//g;
 $dir =~ s/\/+$//;
 my @dirs = ($dir);
 my @sammel;
 while (my $xdir = pop(@dirs))
 {
  push(@sammel,$xdir);
  push(@dirs,opendir_pathdirs($xdir));
 }
 return @sammel;
}
sub opendir_subpathfiles($)
{
 my $dir = $_[0];
 $dir =~ s/\\/\//g;
 $dir =~ s/\/+$//;
 my @dirs = ($dir);
 my @sammel;
 while (my $xdir = pop(@dirs))
 {
  push(@sammel,opendir_pathfiles($xdir));
  push(@dirs,opendir_pathdirs($xdir));
 }
 return @sammel;
}
sub readfile($)
{
 if(open RIN,"<$_[0]") { my $old=$/; undef $/; my $content=<RIN>; $/=$old; close RIN; return($content); }  return "";
}
sub readfile_binmode($)
{
 if(open RIN,"<$_[0]") { binmode RIN; my $old=$/; undef $/; my $content=<RIN>; $/=$old; close RIN; return($content); }  return "";
}
sub readfile_array($)
{
 if(open RIN,"<$_[0]") { my @content=<RIN>; close RIN; return(@content); }  return "";
}
sub readfile_array_nobr($)
{
 if(open RIN,"<$_[0]") { my $old=$/; undef $/; my $content=<RIN>; $/=$old; close RIN; return(split("\n",$content)); }  return "";
}
sub writefile($$) 
{
 if(open WOUT,">$_[0]") { print WOUT $_[1]; close WOUT; }
}
sub writefile_binmode($$) 
{
 if(open WOUT,">$_[0]") { binmode WOUT; print WOUT $_[1]; close WOUT; }
}
sub writefile_binmode_utf8($$) 
{
 if(open WOUT,">$_[0]") { binmode WOUT,":utf8"; print WOUT $_[1]; close WOUT; }
}
sub writefile_append($$) 
{
 if(open WOUT,">>$_[0]") { print WOUT $_[1]; close WOUT; }
}
sub writefile_appendbinmode($$) 
{
 if(open WOUT,">>$_[0]") { binmode WOUT; print WOUT $_[1]; close WOUT; }
}
sub print_ub($) #unbuffered
{
 my $old = $|; $| = 1; print $_[0]; $| = $old;
}

#==== Webload SUBS
sub webload(@) 
{
 # example: my ($content,$base,$code,$status,$ct) = std::webload($href_filled,$browser_names[int(rand(117)],"","silent",400000,"",10);
 my $link = $_[0];
 my $agent = $_[1];
 my $proxy = $_[2];
 my $silent = $_[3];
 my $max_size = $_[4];
 my $accept_decodable = $_[5];
 my $timeout = $_[6];

 #---- ua
 #my $ua = LWP::UserAgent->new;
 my $ua = LWP::UserAgent->new(ssl_opts => {verify_hostname => 0,timeout => (($timeout ne "")?$timeout:100),Timeout => (($timeout ne "")?$timeout:100)});
 $ua->agent(($agent ne "")?$agent:"Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1;)");
 $ua->max_redirect(4);
 $ua->timeout($timeout) if ($timeout ne "");
 $ua->max_size($max_size) if ($max_size ne "");
 $ua->proxy(['http','ftp'], $proxy) if ($proxy ne ""); #'http://user:pass@world.proxymesh.com:31280/'
 
 #---- load
 my $request = HTTP::Request->new(GET => $link);
 $request->accept_decodable() if ($accept_decodable ne "no");
 my $response = $ua->request($request);
 $response->is_success;
 my $data = $response->decoded_content;
 print "Download ".$response->code." ".length($data).": $link\n" if ($silent ne "silent");
 return($data,$response->base,$response->code,$response->status_line,$response->header("Content-Type"));
}

#======================================= STD COMPRESS SUBS
#----------------------- mem(de)compress
sub memcompress_zlib($) 
{
 return Compress::Zlib::memGzip($_[0]);#type1
 #return compress($_[0],9);#type2
}
sub memdecompress_zlib($) 
{
 return Compress::Zlib::memGunzip($_[0]);
}
#----------------------- writefile
sub writefile_bzip2($$) 
{
 my $bz = Compress::Bzip2->new( -blockSize100k => 1, -verbosity => 0 );
 $bz->bzopen($_[0], 'w');
 #$byteswritten = $bz->bzwrite($_[1]);
 $bz->bzwrite($_[1]);
}
sub writefile_zlib($$) 
{
 #---- type1
 #$dest = Compress::Zlib::memGzip($_[1]);
 #writefilebinmode($_[0],$dest);
 #---- type2
 #if(open WOUT,">$_[0]")
 #{
 # binmode WOUT;
 # my $dest = compress($_[1],9);
 # print WOUT $dest;
 # close WOUT;
 #}
 #---- type3
 if(open WOUT,">$_[0]")
 {
  binmode WOUT;
  #print $_[0]."\n";
  my $gz = gzopen(\*WOUT, "wb9") or warn "Cannot open wout: $gzerrno\n";
  $gz->gzwrite($_[1]) or warn "error writing: $gzerrno\n";
  #print WOUT $dest;
  $gz->gzclose ;
  close WOUT;
 }
}
#----------------------- readfile
sub readfile_bzip2($) 
{
 my $bz = Compress::Bzip2->new( -blockSize100k => 1, -verbosity => 0 );
 $bz->bzopen($_[0], 'r');
 my $buff;
 $bz->bzread($buff,999999999);# sorry, must set something
 return $buff;
}
sub readfile_zlib($) 
{
 if(open WIN,"<$_[0]")
 {
  binmode WIN;
  #print $_[0]."\n";
  my $gz = gzopen(\*WIN, "rb") or warn "Cannot open wout: $gzerrno\n";
  my $buff;
  $gz->gzread($buff,999999999) or warn "error writing: $gzerrno\n";
  #print WIN $dest;
  $gz->gzclose;
  close WIN;
  return $buff;
 }
}

#=========================== Global Messageboxes
# example: return if std::quesmessbox(hex($mw->id),"");
sub infomessbox($$)
{
 Win32::GUI::MessageBox($_[0],$_[1],'Job Join',0x11040);
 return 1;
}
sub quesmessbox($$)
{
 my $answer = Win32::GUI::MessageBox($_[0],$_[1],'Job Join',0x44);
 return ($answer == 6)? 1: 0;
}

#==== Regex as Function
sub rx(@)
{
 my $str = $_[0];
 $str =~ s/$_[1]/$_[2]/;
 return $str;
}
sub rx_g(@)
{
 my $str = $_[0];
 $str =~ s/$_[1]/$_[2]/g;
 return $str;
}

#========================================= my read write plan
#---- mymkdir
sub mymkdir($)
{
 my $fullpath = shift;
 if (! -e $fullpath)
 {
  my @pathsplit = split(/[\/\\]+/,$fullpath);
  my $sammelpath = "";
  for(my $i = 0; $i <= $#pathsplit-1; $i++)
  {
   $sammelpath .= $pathsplit[$i]."/";
   #print($sammelpath) if (! -d $sammelpath);
   mkdir($sammelpath) if (! -d $sammelpath);

  }
 }
 return($fullpath);
}
#perl -e "use std; std::mymkdir2('xxx/ff/ff/');"
#sub mymkdir2($)
#{
# my $fullpath = $_[0];
# $fullpath =~ s/[^\/]*$//;
# return if ($fullpath eq '');
# if (! -d $fullpath)
# {
#  my @pathsplit = split('/',$fullpath);
#  for(my $i = $#pathsplit; $i > 0; $i--)
#  {
#my @pathsplitx = @pathsplit;
#splice(@pathsplitx,$i);
#print join('/',@pathsplitx)."\n";
#print $i."\n";
#
#  }
#
# }
#}
#---- xwrite
sub xwrite($$@)
{
 my $order  = shift;
 my $wurzel = shift;
 my @inhalt = @_;

#print "size: ".$#inhalt."\n";
#print "inhalt0: ".$inhalt[0]."\n";
#print "0-not-defined\n" if (not defined $inhalt[0]);

 $wurzel =~ s/([~\t\n])/"~[".ord($1)."]"/ge;
 #$inhalt =~ s/([~\t\n])/"~[".ord($1)."]"/ge;
 map($_ =~ s/([~\t\n,])/"~[".ord($1)."]"/ge,@inhalt);

 #my $path_and_file = $order."/".substr(sha1_hex($wurzel),0,2)."/".substr(sha1_hex($wurzel),2,2)."/".substr(sha1_hex($wurzel),4,2);
 #my $path_and_file = "equals/".$order."/".substr(sha1_hex($wurzel),0,2);
 #my $path_and_file = "equals/".$order."/".sprintf("%03s",0xFFFF/hex(substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,4))*999);
 my $path_and_file = "equals/".$order."/".sprintf("%03s",int(hex(substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,4))/0xFFFF*999)).".txt";
 #my $path_and_file = "equals/".$order."/test.txt";
 mymkdir($path_and_file);

 my $found = 0;
 my $sammelfile = "";
 my $old = $|; $| = 1; # gepufferte ausgabe abschalten damit perl nicht noch schreibt und flock schon aufgehoben ist
 #open(DATEI, "<$path_and_file");
 sysopen(DATEI, $path_and_file, O_RDWR | O_CREAT);
 flock(DATEI, LOCK_EX);
 while(<DATEI>)
 {
  my $tabindex = index($_,"\t");
  if (substr($_,0,$tabindex) ne $wurzel) # substr ist schneller als split
  {
   $sammelfile .= $_;
  }
  else
  {
   if (substr($_,$tabindex+1,-2) eq join(",",@inhalt)) # schreiben verhindern wenn gleich
   { $| = $old; close(DATEI); return; }
   $found = 1;
  }
 }
 #close(DATEI);

 if ($found == 0)
 {
  # schreiben verhindern wenn undef und eh nicht in file
  if ($#inhalt <= 0 && not defined $inhalt[0])
  { $| = $old; close(DATEI); return; }
  # anhaengendes schreiben wenn defined und war nicht in file
  #open(DATEI, ">>$path_and_file");
  print DATEI "$wurzel\t".join(",",@inhalt)."\t\n";
  close(DATEI);
 }
 else
 {
   #open(DATEI, ">$path_and_file");
   sysseek(DATEI, 0, 0);
   truncate (DATEI, 0);
   print DATEI (($#inhalt <= 0 && not defined $inhalt[0])? "": "$wurzel\t".join(",",@inhalt)."\t\n").$sammelfile;
   #print "l:".length($newcontent)."\n";
   close(DATEI);
 }
}

#---- xwritei
sub xwritei($$@)
{
 my $order   = shift;
 my $wurzel  = shift;
 my $wurzeli = lc($wurzel);
 my @inhalt  = @_;

 $wurzel =~ s/([~\t\n])/"~[".ord($1)."]"/ge;
 #$inhalt =~ s/([~\t\n])/"~[".ord($1)."]"/ge;
 map($_ =~ s/([~\t\n,])/"~[".ord($1)."]"/ge,@inhalt);

 #my $path_and_file = $order."/".substr(sha1_hex($wurzel),0,2)."/".substr(sha1_hex($wurzel),2,2)."/".substr(sha1_hex($wurzel),4,2);
 #my $path_and_file = "equals/".$order."/".substr(sha1_hex($wurzel),0,2);
 #my $path_and_file = "equals/".$order."/".sprintf("%03s",0xFFFF/hex(substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,4))*999);
 my $path_and_file = "equals/".$order."/".sprintf("%03s",int(hex(substr(Digest::SHA->new(1)->add($wurzeli)->hexdigest,0,4))/0xFFFF*999)).".txt";
 mymkdir($path_and_file);

 my $sammelfile = "";
 open(DATEI, "<$path_and_file");
 while(<DATEI>)
 {
  $sammelfile .= $_ if (lc(substr($_,0,index($_,"\t"))) ne $wurzeli); # substr ist schneller als split
 }
 close(DATEI);

 #$sammelfile .= "$wurzel\t".join(",",@inhalt)."\t\n";

 open(DATEI, ">$path_and_file");
 print DATEI (($#inhalt != -1)? "$wurzel\t".join(",",@inhalt)."\t\n": "").$sammelfile;
 close(DATEI);
}

#---- xwritecount
sub xwritecount($$$)
{
 my $order  = $_[0];
 my $wurzel = $_[1];
 my $count  = $_[2];

 $wurzel =~ s/([~\t\n])/"~[".ord($1)."]"/ge;

 #my $path_and_file = "equals/".$order."/".sprintf("%03s",0xFFFF/hex(substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,4))*999);
 my $path_and_file = "equals/".$order."/".sprintf("%03s",int(hex(substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,4))/0xFFFF*999)).".txt";
 mymkdir($path_and_file);

 my $sammelfile = "";
 my $found = 0;
 sysopen(DATEI, $path_and_file, O_RDWR | O_CREAT);
 flock(DATEI, LOCK_EX);
 while(<DATEI>)
 {
  my ($wurzelx,$inhaltx) = split("\t",$_);
  if ($wurzelx ne $wurzel)
  {
   $sammelfile .= $_;
  }
  else
  {
   $found = 1;
   $sammelfile .= "$wurzelx\t".($inhaltx+$count)."\t\n";
  }
 }

 $sammelfile .= "$wurzel\t".$count."\t\n" if ($found eq 0);

 sysseek(DATEI, 0, 0);
 truncate (DATEI,0);
 print DATEI $sammelfile;
 flock(DATEI, LOCK_UN);
 close(DATEI);
}

#---- xread
sub xread($$)
{
 my $order = $_[0];
 my $wurzel = $_[1];
 $wurzel =~ s/([~\t\n])/"~[".ord($1)."]"/ge;

 #my $path_and_file = $order."/".substr(sha1_hex($wurzel),0,2)."/".substr(sha1_hex($wurzel),2,2)."/".substr(sha1_hex($wurzel),4,2);
 #my $path_and_file = "equals/".$order."/".substr(sha1_hex($wurzel),0,2);
 #my $path_and_file = "equals/".$order."/".substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,2);
 #my $path_and_file = "equals/".$order."/".sprintf("%03s",0xFFFF/hex(substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,4))*999);
 my $path_and_file = "equals/".$order."/".sprintf("%03s",int(hex(substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,4))/0xFFFF*999)).".txt";
 #my $path_and_file = "equals/".$order."/test.txt";

 #my $my_inhalt = "undef";
 #open(DATEI, "<$path_and_file");
 sysopen(DATEI, $path_and_file, O_RDONLY);
 flock(DATEI, LOCK_SH);
 while(<DATEI>)
 {
  #my ($wurzelx,$inhaltx) = split("\t",$_);
  my $tabindex = index($_,"\t");
  if (substr($_,0,$tabindex) eq $wurzel) # substr ist schneller als split
  {
   #$my_inhalt = ($inhaltx eq '')? 'undef' : $inhaltx;
   ##$my_inhalt =~ s/~\[([0-9]+)\]/chr($1)/ge;
   #last;
   my @data = split(",",substr($_,$tabindex+1,-2));
   map($_ =~ s/~\[([0-9]+)\]/chr($1)/ge,@data);
   close(DATEI), return (wantarray)? @data: $data[0];
  }
 }
 close(DATEI);
 #return map($_ =~ s/~\[([0-9]+)\]/chr($1)/ge,split(",",$my_inhalt));
 return;
}

#---- xreadi
sub xreadi($$)
{
 my $order   = $_[0];
 my $wurzel  = $_[1];
 my $wurzeli = lc($wurzel);
 $wurzel =~ s/([~\t\n])/"~[".ord($1)."]"/ge;

 #my $path_and_file = $order."/".substr(sha1_hex($wurzel),0,2)."/".substr(sha1_hex($wurzel),2,2)."/".substr(sha1_hex($wurzel),4,2);
 #my $path_and_file = "equals/".$order."/".substr(sha1_hex($wurzel),0,2);
 #my $path_and_file = "equals/".$order."/".substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,2);
 #my $path_and_file = "equals/".$order."/".sprintf("%03s",0xFFFF/hex(substr(Digest::SHA->new(1)->add($wurzel)->hexdigest,0,4))*999);
 my $path_and_file = "equals/".$order."/".sprintf("%03s",int(hex(substr(Digest::SHA->new(1)->add($wurzeli)->hexdigest,0,4))/0xFFFF*999)).".txt";

 #my $my_inhalt = "undef";
 open(DATEI, "<$path_and_file");
 while(<DATEI>)
 {
  #my ($wurzelx,$inhaltx) = split("\t",$_);
  my $tabindex = index($_,"\t");
  if (lc(substr($_,0,$tabindex)) eq $wurzeli) # substr ist schneller als split
  {
   #$my_inhalt = ($inhaltx eq '')? 'undef' : $inhaltx;
   ##$my_inhalt =~ s/~\[([0-9]+)\]/chr($1)/ge;
   #last;
   my @data = split(",",substr($_,$tabindex+1,-2));
   map($_ =~ s/~\[([0-9]+)\]/chr($1)/ge,@data);
   close(DATEI), return (wantarray)? @data: $data[0];
  }
 }
 close(DATEI);
 #return map($_ =~ s/~\[([0-9]+)\]/chr($1)/ge,split(",",$my_inhalt));
 return;
}

#==== nobr: zum loeschen aller breaks (\n) per funktion
sub nobr($)
{
 my $stringtobreak = $_[0];
 $stringtobreak =~ s/\n//g;
 return($stringtobreak);
}

#==== (un)fileable
sub fileable_underlineall($)
{
 my $string = $_[0];
 $string =~ s/([~\\\/:\*\?"<>|\n\s\t])/_/g;
 return $string;
}
sub fileable_spaceok($)
{
 my $string = $_[0];
 $string =~ s/([~\\\/:\*\?"<>|\n\t])/"~[".ord($1)."]"/ge;
 return $string;
}
sub fileable($)
{
 my $string = $_[0];
 $string =~ s/([~\\\/:\*\?"<>|\n\s\t])/"~[".ord($1)."]"/ge;
 return $string;
}
sub unfileable($)
{
 my $string = $_[0];
 $string =~ s/~\[([0-9]+)\]/chr($1)/ge;
 return $string;
}

#==== kommas
# gibt nur nachkommateil zurueck
sub kommas($)
{
 return($_[0]-int($_[0]));
}

#============================================== gethashpos
# erstellt einen hashteil
# der hash kann unendlich lang sein
# gibt eine gewisse hashposition zurueck
sub gethashpos($$$)
{
 my $string   = $_[0]; # string to hash
 my $hashpos  = $_[1]; # position im hash
 my $pos_size = $_[2]; # groesse eines hash elements (2=0,1)

 my $posval = 0;
 for(my $pos=0; $pos<length($string); $pos++)
 {
  $posval = ($posval + int(kommas((sin(ord(substr($string,$pos,1)) + $hashpos + $pos + 1) + 1) * 1000) * $pos_size)) % $pos_size;
 }
 return $posval;
}

#============================================== (un)crypt
sub crypt($$)
{
 my $salt = $_[0];
 my $text = $_[1];
 #my @saltsplit = map(ord($_),split('',$salt));
 #my @peppersplit; for(my $i=0; $i<20; $i++) { $peppersplit[$i]=int(rand(128)); }
 my $pepper = join('',map(chr int rand 128,1..20));
 print "pepper: $pepper\n";

 my $sammel;
 for(my $i=0; $i<length($text); $i++)
 {
  #my $pos_salted_peppered = $i + $saltsplit[$i%($#saltsplit+1)] + $peppersplit[$i%($#peppersplit+1)];
  #my $quirk = int(kommas((sin($pos_salted_peppered)+1)*1000) * 128);
  #print "xxx: ".int(($quirk-int($quirk))*1000).", ".$quirk.", ".($i + ($saltsplit[$i % ($#saltsplit+1)]))."\n";
  #$sammel .= chr((ord(substr($text,$i,1)) + $quirk) % 128);
  my $quirk_pos_salted_peppered = gethashpos($salt.$pepper,$i,128);
  $sammel .= chr((ord(substr($text,$i,1)) + $quirk_pos_salted_peppered) % 128);
 }
 #return join('',map(chr($_),@peppersplit)).$sammel;
 return $pepper.$sammel;
}
sub uncrypt($$)
{
 my $salt = $_[0];
 my $text = substr($_[1],20);
 #my @saltsplit   = map(ord($_),split('',$salt));
 #my @peppersplit = map(ord($_),split('',substr($_[1],0,20)));
 my $pepper = substr($_[1],0,20);

 my $sammel;
 for(my $i=0; $i<length($text); $i++)
 {
  #my $pos_salted_peppered = $i + $saltsplit[$i%($#saltsplit+1)] + $peppersplit[$i%($#peppersplit+1)];
  #my $quirk = int(kommas((sin($pos_salted_peppered)+1)*1000) * 128);
  #$sammel .= chr((ord(substr($text,$i,1)) - $quirk) % 128);
  my $quirk_pos_salted_peppered = gethashpos($salt.$pepper,$i,128);
  $sammel .= chr((ord(substr($text,$i,1)) - $quirk_pos_salted_peppered) % 128);
 }
 return $sammel;
}
sub zcrypt($$)   { return crypt($_[0],Compress::Zlib::memGzip($_[1])); } 
sub zuncrypt($$) { return Compress::Zlib::memGunzip(uncrypt($_[0],$_[1])); }
sub zcrypt_hex($$)   { return unpack("H*",zcrypt($_[0],$_[1])); } 
sub zuncrypt_hex($$) { return zuncrypt($_[0],pack("H*",$_[1])); }

#========================================================== tree
# example: my ($state,$wurzel,$data) = std::tree_find('10-10-md',$folder,$wurzel);
# md = migrate data .. means when nothing found get nearest base data
my $sysopen = 1;
#---------------------------------------------- tree_write
sub tree_write($$$$)
{
 my ($size_wurzel,$size_data,$breaksize) = split('-',$_[0]);
 my $parts_size = $size_wurzel+$size_data+3;
 my $folder = $_[1];
 #my $wurzel = substr($_[2],0,$size_wurzel);
 #my $data   = substr($_[3],0,$size_data);
 my $wurzel = (length($_[2]) > $size_wurzel)? substr($_[2],0,$size_wurzel): $_[2];
 my $data   = (length($_[3]) > $size_data)?   substr($_[3],0,$size_data):   $_[3];
 my $file   = "trees/".$folder."/tree.txt";

 mymkdir($file);

 my ($state,$index,$string) = leaf_find($size_wurzel.'-0',$file,$wurzel);
 #my $leaf_id = sha1_hex($string);
 my $leaf_id = Digest::SHA->new(1)->add($string)->hexdigest;
 my $leaf_file = "trees/".$folder."/tree_".$leaf_id.".txt";
 my ($leaf_state,$leaf_index,$leaf_string) = leaf_find($size_wurzel.'-'.$size_data,$leaf_file,$wurzel);
 if ($leaf_state != 1)#($leaf_state ne 1)
 {
  #leaf_write($leaf_file,$wurzel);
  leaf_write_atid($size_wurzel.'-'.$size_data,$leaf_file,$leaf_index,$wurzel,$data);
  my $leaf_size = (-s $leaf_file)/$parts_size;
  if ($leaf_size >= $breaksize)
  {
   my $split_wurzel = leaf_split($size_wurzel.'-'.$size_data,$folder,$leaf_id);
   leaf_write_atid($size_wurzel.'-0',$file,$index,$split_wurzel,'');
  }
 }
 return $leaf_state;
}
#---------------------------------------------- tree_find
sub tree_find($$$)
{
 my ($size_wurzel,$size_data,$migratedata) = split('-',$_[0]);
 my $parts_size               = $size_wurzel+$size_data+3;
 my $folder = $_[1];
 #my $wurzel = substr($_[2],0,$size_wurzel);
 my $wurzel = (length($_[2]) > $size_wurzel)? substr($_[2],0,$size_wurzel): $_[2];
 my $file   = "trees/".$folder."/tree.txt";

 my ($state,$index,$string) = leaf_find($size_wurzel.'-0',$file,$wurzel);
 #return(0,0,'','') if ($state == 0);
 #my $leaf_id = sha1_hex($string);
 my $leaf_id = Digest::SHA->new(1)->add($string)->hexdigest;
 my $leaf_file = "trees/".$folder."/tree_".$leaf_id.".txt";
 return leaf_find($size_wurzel.'-'.$size_data.'-'.$migratedata,$leaf_file,$wurzel);
}
#---------------------------------------------- leaf_split
sub leaf_split($$$)
{
 my ($size_wurzel,$size_data) = split('-',$_[0]);
 my $folder       = $_[1];
 my $leaf_id      = $_[2];
 #my $data         = $_[3];
 my $file         = "trees/".$folder."/tree_".$leaf_id.".txt";
 my $parts_size   = $size_wurzel+$size_data+3;
 my $file_size    = (-s $file);
 my $file_parts   = $file_size/$parts_size;
 my $file_parts_A = int($file_parts/2);
 my $file_parts_B = $file_parts-$file_parts_A;
 my $file_bytes_A = $file_parts_A*$parts_size;
 my $file_bytes_B = $file_parts_B*$parts_size;

 open(DATEI, "+<".$file) || die "leaf_split_open1($file): $!" if ($sysopen == 0);
 sysopen(DATEI, $file, O_RDWR | O_CREAT) || die "leaf_split_open1($file): $!" if ($sysopen == 1);
 seek(DATEI, $file_bytes_A, 0) if ($sysopen == 0);
 sysseek(DATEI, $file_bytes_A, 0) if ($sysopen == 1);
 my $part_B = '';
 read(DATEI, $part_B, $file_bytes_B) if ($sysopen == 0);
 sysread(DATEI, $part_B, $file_bytes_B) if ($sysopen == 1);
 truncate (DATEI,$file_bytes_A);
 close(DATEI);

 my $part_B_headkey = substr($part_B,0,$size_wurzel+1);
 $headkey_wurzel_size = ord(substr($part_B_headkey,0,1));
 $headkey_wurzel_data = substr($part_B_headkey,1,$headkey_wurzel_size);

 #my $file_new = "trees/".$folder."/tree_".sha1_hex($headkey_wurzel_data).".txt";
 my $file_new = "trees/".$folder."/tree_".Digest::SHA->new(1)->add($headkey_wurzel_data)->hexdigest.".txt";

 open(DATEI, ">".$file_new) || die "leaf_split_open2($file): $!" if ($sysopen == 0);
 sysopen(DATEI, $file_new, O_WRONLY | O_CREAT) || die "leaf_split_open2($file): $!" if ($sysopen == 1);
 #print DATEI $part_B if ($sysopen == 0);
 write(DATEI,$part_B) if ($sysopen == 0);
 syswrite(DATEI,$part_B) if ($sysopen == 1);
 close(DATEI);

 #my $file_stamm = "trees/".$folder."/tree.txt";
 #print "stamm: ".$file_stamm."\n";
 return($headkey_wurzel_data);
}
#---------------------------------------------- leaf_write_atid
sub leaf_write_atid($$$$$)
{
 my ($size_wurzel,$size_data) = split('-',$_[0]);
 my $parts_size               = $size_wurzel+$size_data+3;
 my $file          = $_[1];
 my $atid          = $_[2];
 my $atid_byte     = $atid*$parts_size;
 my $wurzel        = (length($_[3]) > $size_wurzel)? substr($_[3],0,$size_wurzel): $_[3];
 my $data          = (length($_[4]) > $size_data)?   substr($_[4],0,$size_data):   $_[4];
 my $length_wurzel = length($wurzel);
 my $length_data   = length($data);

 #open(DATEI, "+>>:crlf",$file) || die "leaf_write_atid_open1: $!";
 #open(DATEI, "+>>".$file) || die "leaf_write_atid_open1: $!";
 sysopen(DATEI, $file, O_RDWR | O_CREAT) || die "leaf_write_atid_sysopen($file): $!";
 sysseek(DATEI, $atid_byte, 0);
 my $part_B = '';
 sysread(DATEI, $part_B, 200000000);
 sysseek(DATEI, $atid_byte, 0);
 syswrite(DATEI,chr($length_wurzel).$wurzel.(' ' x ($size_wurzel-$length_wurzel)).chr($length_data).$data.(' ' x ($size_data-$length_data))."\n".$part_B);
 #print DATEI chr($length_wurzel).$wurzel.(' ' x ($size_wurzel-$length_wurzel)).chr($length_data).$data.(' ' x ($size_data-$length_data))."\n".$part_B;
 #print DATEI chr($size_wurzel).$wurzel.(' ' x ($size_wurzel-length($wurzel))).chr($size_data).$data.(' ' x ($size_data-length($data)))."\n".$part_B;
 #truncate (DATEI,$file_bytes_A);
 close(DATEI);
}
#---------------------------------------------- leaf_write
sub leaf_write($$$$)
{
 my ($size_wurzel,$size_data) = split('-',$_[0]);
 my $file   = $_[1];
 #my $wurzel = substr($_[2],0,$size_wurzel);
 #my $data   = substr($_[3],0,$size_data);
 my $wurzel = (length($_[2]) > $size_wurzel)? substr($_[2],0,$size_wurzel): $_[2];
 my $data   = (length($_[3]) > $size_data)?   substr($_[3],0,$size_data):   $_[3];

 my ($state,$index,$string) = leaf_find($size_wurzel.'-'.$size_data,$file,$wurzel);
print "state:$state\n";
print "index:$index\n";
print "string:$string\n";
print "---------------------\n";
#exit;
 if ($state ne 1)
 {
  leaf_write_atid($size_wurzel.'-'.$size_data,$file,$index,$wurzel,$data);
 }
}
#---------------------------------------------- leaf_check
sub leaf_find($$$)
{
 my ($size_wurzel,
     $size_data,
     $migratedata) = split('-',$_[0]);
 my $file          = $_[1];
 open(tDATEI, "<".$file) || return(0,0,'','') if ($sysopen == 0);
 sysopen(tDATEI, $file, O_RDONLY) || return(0,0,'','') if ($sysopen == 1);
 #my $wurzel            = substr($_[2],0,$size_wurzel);
 my $wurzel            = (length($_[2]) > $size_wurzel)? substr($_[2],0,$size_wurzel): $_[2];
 my $byte              = (-s $file);
 my $parts_size        = $size_wurzel+$size_data+3;
 my $parts             = $byte/$parts_size;
 my $parts_zeroed      = $parts-1;

 my $grenze_kleinst    = 0;
 my $grenze_groesst    = $parts_zeroed;
 my $suchspanne;
 my $suchpunkt;
 my $suchpunkt_byte;
 my $readed_data;
 my $readed_wurzel_size;
 my $readed_wurzel_data;
 my $readed_wurzel_data_lastup = "";
 my $readed_content_size;
 my $readed_content_data;
 while ($grenze_kleinst <= $grenze_groesst)
 {
  $suchspanne        = $grenze_groesst-$grenze_kleinst;
  $suchpunkt         = $grenze_kleinst+int($suchspanne/2);
  $suchpunkt_byte    = $suchpunkt*$parts_size;

  seek(tDATEI, $suchpunkt_byte, 0) if ($sysopen == 0);
  sysseek(tDATEI, $suchpunkt_byte, 0) if ($sysopen == 1);
  read(tDATEI, $readed_data, ($size_wurzel+1)) if ($sysopen == 0);
  sysread(tDATEI, $readed_data, ($size_wurzel+1)) if ($sysopen == 1);
  $readed_wurzel_size = ord(substr($readed_data,0,1));
  $readed_wurzel_data = substr($readed_data,1,$readed_wurzel_size);

  if ($wurzel lt $readed_wurzel_data) # < suche kleiner
  {
   $grenze_groesst = $suchpunkt-1;
  }
  elsif ($wurzel gt $readed_wurzel_data) # > suche grsser
  {
   $grenze_kleinst = $suchpunkt+1;
   $readed_wurzel_data_lastup = $readed_wurzel_data;
  }
  else # = gefunden
  {
   read(tDATEI, $readed_data, ($size_data+1)) if ($sysopen == 0);
   sysread(tDATEI, $readed_data, ($size_data+1)) if ($sysopen == 1);
   $readed_content_size = ord(substr($readed_data,0,1));
   $readed_content_data = substr($readed_data,1,$readed_content_size);
   close(tDATEI);
   return(1,$suchpunkt,$readed_wurzel_data,$readed_content_data);
  }
 }
 $suchpunkt-- if ($wurzel lt $readed_wurzel_data);
 $suchpunkt++;
 if ($suchpunkt eq 0 || $migratedata ne 'md')
 {
  close(tDATEI);
  return(2,$suchpunkt,$readed_wurzel_data_lastup,'');
 }
 else
 {
  seek(tDATEI, ($suchpunkt-1)*$parts_size+$size_wurzel+1, 0) if ($sysopen == 0);
  sysseek(tDATEI, ($suchpunkt-1)*$parts_size+$size_wurzel+1, 0) if ($sysopen == 1);
  read(tDATEI, $readed_data, ($size_data+1)) if ($sysopen == 0);
  sysread(tDATEI, $readed_data, ($size_data+1)) if ($sysopen == 1);
  $readed_content_size = ord(substr($readed_data,0,1));
  $readed_content_data = substr($readed_data,1,$readed_content_size);
  close(tDATEI);
  return(2,$suchpunkt,$readed_wurzel_data_lastup,$readed_content_data);
 }
}
#--------------------------- leaf_list
sub leaf_list($$)
{
 my $size_wurzel = $_[0];# = split('-',$_[0]);
 my $folder      = $_[1];
 my $parts_size  = $size_wurzel+3;

 my $leaf_id = Digest::SHA->new(1)->add('')->hexdigest;
 my $leaf_file = "trees/".$folder."/tree_".$leaf_id.".txt";
 if (! -f "trees/".$folder."/tree.txt")
 {
  #my $leaf_id = Digest::SHA->new(1)->add('')->hexdigest;
  #my $leaf_file = "trees/".$folder."/tree_".$leaf_id.".txt";
  return($leaf_file) if (-f $leaf_file);
  return();
 }
 else
 {
  my @foldersammel = ($leaf_file);
  my $leaf_size = (-s "trees/".$folder."/tree.txt")/$parts_size;
  my $treefile = readfile("trees/".$folder."/tree.txt");
  for(my $i = 0; $i < $leaf_size; $i++)
  {
   my $wurzel = substr($treefile,$i*$parts_size,$size_wurzel+1);
   my $wurzel_size = ord(substr($wurzel,0,1));
   my $wurzel_data = substr($wurzel,1,$wurzel_size);

   my $leaf_id = Digest::SHA->new(1)->add($wurzel_data)->hexdigest;
   push(@foldersammel,"trees/".$folder."/tree_".$leaf_id.".txt");
  }
  return(@foldersammel);
 }
}
#--------------------------- tree_makefile
sub tree_makefile($$)
{
 my ($size_wurzel,$size_data) = split('-',$_[0]);
 my $folder                   = $_[1];
 my $parts_size               = $size_wurzel+$size_data+3;

 writefile("trees/".$folder."/fullfile.txt",'');
 foreach my $file (leaf_list($size_wurzel,$folder))
 {
  print "file: $file\n";
  #my $file              = "trees/".$folder."/tree_".Digest::SHA->new(1)->add($headkey_wurzel_data)->hexdigest.".txt";
  my $byte              = (-s $file);
  my $parts             = $byte/$parts_size;
  #my $parts_zeroed      = $parts-1;
  my $filecontent       = readfile_binmode($file);

  my $sammel = '';
  for(my $i=0; $i<$parts; $i++)
  {
   my $suchpunkt_byte      = $i*$parts_size;
   my $readed_wurzel_size  = ord(substr($filecontent,$suchpunkt_byte,1));
   my $readed_wurzel_data  = substr($filecontent,$suchpunkt_byte+1,$readed_wurzel_size);
   my $readed_content_size = ord(substr($filecontent,$suchpunkt_byte+$size_wurzel+1,1));
   my $readed_content_data = substr($filecontent,$suchpunkt_byte+$size_wurzel+2,$readed_content_size);
   #writefile_append("trees/".$folder."/fullfile.txt",$readed_wurzel_data."\t".$readed_content_data."\n");
   $sammel .= $readed_wurzel_data."\t".$readed_content_data."\n";
  }
  writefile_append("trees/".$folder."/fullfile.txt",$sammel);
 }
}

#############################################################
############################################## zwrite zread #
#############################################################
# Tree gesttzte komprimierte read write funktion.          #
# Daten werden neben dem tree als zdata.dat gespeichert.    #
# Beispiel:                                                 #
# std::zwrite(50,'folder','wurzel','data');                 #
# std::zread('folder','wurzel');                            #
# 50 ist die Leaf groesse vor dem splitten.                 #
# zread gibt 'undef' zurck wenn leer.                      #
#############################################################
sub zwrite($$$$)
{
 my $leafsplitsize     = $_[0];
 my $folder            = $_[1];
 my $wurzel_sha        = Digest::SHA->new(1)->add($_[2])->hexdigest;
 my $data_packed       = memcompress_zlib($_[3]);
 my $data_size_int     = pack('I1',length($data_packed));
 my $datafile          = 'trees/'.$folder.'/zdata.dat';
 my $datafile_size_int = pack('I1',(-s $datafile));

 my $status = tree_write('40-8-'.$leafsplitsize,$folder,$wurzel_sha,$data_size_int.$datafile_size_int);
 writefile_appendbinmode($datafile,$data_packed) if ($status != 1);
 #writefile_appendbinmode('c:\\data.dat',$data_packed);
}
sub zread($$)
{
 my $folder            = $_[0];
 my $wurzel_sha        = Digest::SHA->new(1)->add($_[1])->hexdigest;

 my ($tree_state,$tree_index,$tree_wurzel,$tree_data) = tree_find('40-8',$folder,$wurzel_sha);
 return('undef') if($tree_state != 1);

 my $datafile = 'trees/'.$folder.'/zdata.dat';
 my $data_pos = unpack('I1',substr($tree_data,4,4));
 my $data_len = unpack('I1',substr($tree_data,0,4));

 sysopen(DATEI, $datafile, O_RDONLY);
 sysseek(DATEI, $data_pos, 0);
 my $data;
 sysread(DATEI, $data, $data_len);
 close(DATEI);

 return memdecompress_zlib($data);
}





1;
